
# C3 password interactions with user and/or environment variables for private keys

import sys, os

from c3.constants import PASS_VAR, SHOW_VAR
from c3.errors import NoPassword

if sys.platform == "win32":
    from msvcrt import getch  # noqa
else:
    import tty, termios
    # Note: this getch fails if not isatty on macos, and just ignores stdin on windows.
    def getch():
        fd = sys.stdin.fileno()
        old_settings = termios.tcgetattr(fd)
        try:
            tty.setraw(sys.stdin.fileno())
            ch = sys.stdin.read(1)
        finally:
            termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
        return ch


# Policy: we're not supporting stdin-redirect for entering passwords.
#         it's environment variable or interactive entry only.
# Policy: environment variable password must not be blank if present.

def get_env_password():
    if PASS_VAR and PASS_VAR in os.environ and os.environ[PASS_VAR]:
        return os.environ[PASS_VAR]
    return ""


def get_enter_password(prompt):
    if not sys.stdin.isatty():       # this assumes get_env_password has been called
        raise ValueError("Password can't be entered (%r environment variable can be used)" % PASS_VAR)
    else:
        return enter_password(prompt, SHOW_VAR not in os.environ)


def enter_password(prompt="Password: ", mask=True):
    sys.stdout.write(prompt)
    sys.stdout.flush()
    pw = []
    while True:
        cc = ord(getch())
        if cc == 13:            # enter
            sys.stdout.write("\n")
            sys.stdout.flush()
            return "".join(pw)
        elif cc == 27:       # escape
            raise NoPassword("No password supplied")
        elif cc == 3:        # ctrl=c
            raise KeyboardInterrupt
        elif cc in (8, 127):  # backspace, del
            if len(pw) > 0:
                sys.stdout.write("\b \b")
                sys.stdout.flush()
                pw = pw[:-1]
        elif 0 <= cc <= 31:    # unprintables
            pass
        else:               # add to password
            if mask:
                sys.stdout.write("*")
            else:
                sys.stdout.write(chr(cc))
            sys.stdout.flush()
            pw.append(chr(cc))

# This is used when SETTING a password.
# It fetches from env if present,
# only does a single enter if masking is OFF,
# and does the "enter twice" if masking is on.
# Note: (because this is for setting, we can have the loop in here, unlike with getting, where the
#        loop has to be synchronous with trying to decrypt the thing, so has to be in caller's code.)

def get_double_enter_setting_password(prompt1, prompt2):
    # --- Try password from environment variables ---
    passw = get_env_password()
    if passw:
        return passw

    # --- Get password from user once because showing the characters ---
    if SHOW_VAR in os.environ:        # dont do an enter-re-enter if show pass is on
        pass1 = get_enter_password(prompt1)
        return pass1
    else:
        while True:
            pass1 = get_enter_password(prompt1)
            pass2 = get_enter_password(prompt2)
            if pass1 == pass2:
                return pass1
            else:
                print("Sorry, entered passwords do not match, please try again")
                continue


# Testing is fairly manual, for obj reasons, something like:
# python -c "import getpassword ; print('Pass: ',getpassword.get_password('>: ', 'PASS_VAR', 'PASS_SHOW'))"

if __name__ == '__main__':
    pw = get_enter_password()
    print("Password got:  ",repr(pw))





# C3 data schema for the binary blocks, tag constants, and API enums

import b3

# --- API Actions ---

# Make/Sign actions:
MAKE_SELFSIGNED = 1
MAKE_INTERMEDIATE = 2
SIGN_PAYLOAD = 3
LINK_APPEND = 1
LINK_NAME = 2

# --- environment variables for priv key crypting ---
PASS_VAR = "C3_PASSWORD"
SHOW_VAR = "C3_SHOW_PASS"


# --- Top-level tag values ---
PUB_CSR = 10
DUALBLOCK = 11
PUB_PAYLOAD = 12      # cert chain with a payload as the first entry
BARE_PAYLOAD = 13     # literally just payload bytes but tagged with a header tag.
PUB_CERTCHAIN = 14    # cert chain with a cert as the first entry
PRIV_CRCWRAPPED = 15  # "priv data with a crc32 integrity check"

# Public-part chain-level
HDR_DAS = 0x19            # "data_part and sig_part structure"


# Private-part field types
PRIVTYPE_BARE = 1
PRIVTYPE_PASS_PROTECT = 2
KT_ECDSA_PRIME256V1 = 1
KT_ECDSA_SECP256K1 = 2      # the bitcoin one, for which fast implementations exist
KNOWN_PRIVTYPES = [1, 2]
KNOWN_KEYTYPES = [1]


# --- Public-Part data structures ---

CERT_SCHEMA = (
    (b3.BYTES,     "cert_id",       0, True),
    (b3.UTF8,      "subject_name",  1, True),
    (b3.UVARINT,   "key_type",      2, True),
    (b3.BYTES,     "public_key",    3, True),
    (b3.BASICDATE, "expiry_date",   4, True),
    (b3.BASICDATE, "issued_date",   5, True),
    (b3.UTF8,      "cert_type",     6, False),
)

SIG_SCHEMA = (
    (b3.BYTES, "signature", 0,  True),
    (b3.BYTES, "signing_cert_id", 1, False),  # value can be empty.
)

DATASIG_SCHEMA = (
    (b3.BYTES, "data_part", 0, True),  # a cert (CERT_SCHEMA) or a payload (BYTES)
    (b3.BYTES, "sig_part", 1, True),   # a SIG_SCHEMA
    # (We could put a sig_list item here later if we want to go chain multi sig.)
)


# --- Private block top level data structures ---

PRIV_CRCWRAP_SCHEMA = (
    (b3.UVARINT, "priv_type", 0, True),      # protection method (e.g. bare/none, or pass_protect)
    (b3.UVARINT, "key_type",  1, True),      # actual type of private key (e.g. ecdsa 256p)
    (b3.BYTES,   "priv_data", 2, True),
    (b3.UVARINT, "crc32",     3, True),       # crc of privdata for integrity check
)

# --- Both-parts binary data structures ---

DUALBLOCK_SCHEMA = (
    (b3.BYTES, "public", 0, True),
    (b3.BYTES, "private", 1, True),
)



KEY2NAME = {55 : "PUB_PAYLOAD", 66 : "PUB_CERTCHAIN", 77 : "HDR_DAS", 88 : "PRIV_CRCWRAPPED"}




# C3 errors / Exceptions

class C3Error(ValueError):
    pass
class StructureError(C3Error):  # something wrong with the data/binary structure; misparse, corrupt
    pass
class IntegrityError(StructureError):  # the crc32 in the privkey block doesn't match block contents
    pass
class TextStructureError(StructureError):        # something wrong with text structure
    pass
class VerifyError(C3Error):     # parent error for failures in the verification process
    pass
class InvalidSignatureError(VerifyError):   # cryptographic signature failed verification
    pass
class CertNotFoundError(VerifyError):   # chain points to a cert name we dont have in Trusted
    pass
class ShortChainError(VerifyError):  # the next cert for verifying is missing off the end
    pass
class UntrustedChainError(VerifyError):  # the chain ends with a self-sign we dont have in Trusted
    pass
class TamperError(VerifyError):     # visible Fields are present in the textual file,
    pass                             #   but don't match up with the secure fields
class SignError(C3Error):
    pass
class OutputError(C3Error):         # cant output a CE for some reason - usually bc private key not encrypted
    pass
class CertExpired(SignError):       # can't sign, --using's cert has expired.
    pass
class NoPassword(C3Error):          # user aborted entering passwords
    pass





import weakref

from c3.constants import *
from c3.errors import *
from c3 import textfiles
from c3 import structure
from c3 import getpassword

from c3.structure import AttrDict

# --- Data output classes ---
# These are attached to the CEs as properties .pub. .priv. .both.  so e.g. ce.both.as_text()
# so we can go e.g. ce.pub.as_text(),  ce.both.as_binary(),

# Note: textfiles.load_files is the inverse of the write_text_file functions here.

# Policy: It is expected with these out objects that if a priv one is called (priv or both), that
#         priv is wanted. So e.g. .both. should fail if no priv present, rather than being smart.

class CeOutPub(object):
    def __init__(self, ce_parent):
        self.ce = ce_parent

    def as_binary(self):
        return self.ce.pub_block

    def as_text(self, vis_map=None, desc=""):
        return textfiles.make_pub_txt_str_ce(self.ce, desc, vis_map)

    def write_text_file(self, filename, vis_map=None, desc=""):
        txt = self.as_text(vis_map, desc)
        fname = filename + ".public.b64.txt"
        with open(fname, "wt") as f:
            f.write(txt)
        if self.ce.write_prints:
            print("Wrote file '%s'" % (fname,))
        return


class CeOutPriv(object):
    def __init__(self, ce_parent):
        self.ce = ce_parent

    def as_binary(self):
        if not self.ce.epriv_block:
            raise OutputError("Please encrypt() or nopassword() the private key")
        return self.ce.epriv_block

    def as_text(self, vis_map=None, desc=""):
        if not self.ce.epriv_block:
            raise OutputError("Please encrypt() or nopassword() the private key")
        return textfiles.make_priv_txt_str_ce(self.ce, desc)  # note vis_map not used

    def write_text_file(self, filename, vis_map=None, desc=""):
        txt = self.as_text(vis_map, desc)
        fname = filename + ".PRIVATE.b64.txt"
        with open(fname, "wt") as f:
            f.write(txt)
        if self.ce.write_prints:
            print("Wrote file '%s'" % (fname,))


class CeOutBoth(object):
    def __init__(self, ce_parent):
        self.ce = ce_parent

    def as_binary(self):
        if not self.ce.epriv_block:
            raise OutputError("Please encrypt() or nopassword() the private key")
        return structure.combine_binary_pub_priv(self.ce.pub_block, self.ce.epriv_block)

    def as_text(self, vis_map=None, desc=""):
        if not self.ce.epriv_block:
            raise OutputError("Please encrypt() or nopassword() the private key")
        pub_str = textfiles.make_pub_txt_str_ce(self.ce, desc, vis_map)
        priv_str = textfiles.make_priv_txt_str_ce(self.ce, desc)
        return "\n" + pub_str + "\n" + priv_str + "\n"

    def write_text_file(self, filename, vis_map=None, desc=""):
        txt = self.as_text(vis_map, desc)
        fname = filename+".b64.txt"
        with open(fname, "wt") as f:
            f.write(txt)
        if self.ce.write_prints:
            print("Wrote file '%s'" % (fname,))




class CertEntry(object):
    def __init__(self, parent):
        self.parent = weakref.ref(parent)   # pointer to the SignVerify object whose registry(s) we live in
        self.pub_type = 0                   # tag value e.g. PUB_CSR
        self.name = ""

        self.pub_text = ""
        self.epriv_text = ""

        self.pub_block = b""
        self.epriv_block = b""          # bytes of packed PRIV_CRCWRAPPED structure

        self.priv_d = {}                # unpacked PRIV_CRCWRAPPED structure
        self.priv_key_bytes = b""       # actual private key bytes (priv_d.priv_data)

        self.payload = b""              # this or cert, depending on pub_type
        self.cert = {}                  # aka chain[0]
        self.sig = b""

        self.chain = []
        self.vcerts = []                # generated by verify, computed certchain, includes trusted.

        self.vis_map = {}
        self.default_vismap = dict(schema=CERT_SCHEMA,
                                   field_map=["subject_name", "expiry_date", "issued_date", "cert_type"])

        self.files_combined = True      # whether cert was loaded from split files or combined file
        self.write_prints = False       # print helpful messages to console when writing files

        # Output class instances, so user can go ce.pub.as_text(), ce.both.as_binary() etc.
        self.pub = CeOutPub(self)
        self.priv = CeOutPriv(self)
        self.both = CeOutBoth(self)


    # ============== Metadata access functions =============================================

    def vchain(self):
        if len(self.vcerts) > 1 and self.vcerts[0] == self.vcerts[1]:
            return self.vcerts[1:]      # verify() can double-add the root cert under certain
        return self.vcerts              # circumstances (link name vs append etc)

    def blankify(self, dx, key):        # turns absence and None into "<none>"
        if key not in dx:
            return "<none>"
        if dx[key] is None:
            return "<none>"
        return str(dx[key])

    def vtypes(self):
        return "/"+"/".join([self.blankify(i, "cert_type") for i in self.vchain()])

    def vnames(self):
        return "/"+"/".join([self.blankify(i, "subject_name") for i in self.vchain()])


    # ============== Private key encrypt ===================================================

    def private_key_encrypt_sanity_checks(self):
        if not self.priv_key_bytes:
            raise ValueError("CE has no private key bytes")

    def private_key_encrypt(self, password):
        self.private_key_encrypt_sanity_checks()
        epriv_bytes = self.parent().pass_protect.SinglePassEncrypt(self.priv_key_bytes, password)
        self.epriv_block = structure.make_priv_block(epriv_bytes, bare=False)
        return

    def private_key_encrypt_user(self):
        self.private_key_encrypt_sanity_checks()
        prompt1 = "Enter password to set on private key > "
        prompt2 = "Re-enter set private key password    > "
        passw = getpassword.get_double_enter_setting_password(prompt1, prompt2)
        epriv_bytes = self.parent().pass_protect.SinglePassEncrypt(self.priv_key_bytes, passw)
        self.epriv_block = structure.make_priv_block(epriv_bytes, bare=False)

    def private_key_set_nopassword(self):
        self.private_key_encrypt_sanity_checks()
        self.epriv_block = structure.make_priv_block(self.priv_key_bytes, bare=True)

    # ============== Private key decrypt ===================================================

    # out: True - dealt with it (BARE), False= password still needs doing.
    def private_key_decrypt_sanity_checks(self):
        # already have priv_d from load() calling structure.load_priv_block
        if self.priv_d.priv_type == PRIVTYPE_BARE:    # just in case the user calls us when they're not
            self.priv_key_bytes = self.priv_d.priv_data  # supposed to (e.g. a bare key was loaded)
            return True   # be good about it anyway so the user doesn't have to special case things.

        if self.priv_d.priv_type != PRIVTYPE_PASS_PROTECT:
            raise StructureError("Private key is encrypted with an unknown encryption")
        # todo: we dont support this here yet
        if self.parent().pass_protect.DualPasswordsNeeded(self.priv_d.priv_data):
            raise NotImplementedError("Private key wants dual passwords")
        return False

    def private_key_decrypt(self, password):
        if self.private_key_decrypt_sanity_checks():    # true = we're done (e.g BARE)
            return
        priv_ret = self.parent().pass_protect.SinglePassDecrypt(self.priv_d.priv_data, password)
        self.priv_key_bytes = priv_ret

    def private_key_decrypt_user(self):
        if self.private_key_decrypt_sanity_checks():     # true = we're done (e.g BARE)
            return
        # --- Try password from environment variables ---
        passw = getpassword.get_env_password()
        if passw:
            priv_ret = self.parent().pass_protect.SinglePassDecrypt(self.priv_d.priv_data, passw)
            # Note exceptions are propagated right out here, its an exit if this decrypt fails.
            self.priv_key_bytes = priv_ret
            return

        # --- Try password from user ---
        prompt = "Password to unlock private key > "
        priv_ret = b""
        while not priv_ret:
            passw = getpassword.get_enter_password(prompt)
            try:
                priv_ret = self.parent().pass_protect.SinglePassDecrypt(self.priv_d.priv_data, passw)
            except Exception as e:
                print("Failed decrypting private key: ", str(e))
                continue        # let user try again
        self.priv_key_bytes = priv_ret
        return

    # Note: there is no inverse for private_key_nopassword because
    #       incoming BARE keys are handled directly by load()

    def private_key_unload(self):
        self.priv_key_bytes = b""


    # ============== Files output ===========================================================

    # Note: For use by commandline operations.
    # text file writer, partial inverse of signverify.load().
    def write_files(self, parts=None, write_prints=False):
        self.write_prints = write_prints
        # --- Caller has specified --parts explicitly, priv MUST exist. ---
        #    (priv & both will error if not)
        if parts:
            if parts == "split":
                self.pub.write_text_file(self.name)
                self.priv.write_text_file(self.name)
            elif parts == "combine":
                self.both.write_text_file(self.name)
            else:
                raise OutputError("--parts= must be 'split' or 'combine'")
            return
        # --- Special case - after make_csr if there's priv_key_bytes but no epriv_block, ---
        # --- error out because user needs to call encrypt or nopass.                     ---
        if self.priv_key_bytes and not self.epriv_block:
            raise OutputError("Please encrypt() or nopassword() the private key")
        # --- Otherwise if there's no epriv_block (e.g. pub-only load) just write pub ---
        if not self.epriv_block:
            self.pub.write_text_file(self.name)
            return
        # --- If there IS epriv_block then save it how it was loaded ---
        if self.files_combined:
            self.both.write_text_file(self.name)
        else:
            self.pub.write_text_file(self.name)
            self.priv.write_text_file(self.name)




# This is used by C3 privcrypt, but is seperate and distinct code currently

import sys, os, ctypes, itertools, logging, platform
try:
    from pathlib import Path
except ImportError:
    from pathlib2 import Path

from b3 import encode_uvarint, decode_uvarint

# Password protection for private keys (or whatever else)
# WE are payload agnostic - dont care about key_types etc - all that stuff is the Caller's problem
# We're stealing most of this from qsafiles socrypt.

# Note: if libsodium load fails on startup, hold that error and raise it in the user functions.
#       so that libsodium not loading only becomes an error when users try and encrypt/decrypt later.


# Binary format: [salt sz][xpayload sz][salt][num-passes][xpayload][xpayload][xpayload]...

class DecryptFailed(Exception):     pass        # can interpret as Incorrect Password

class PassProtect(object):
    def __init__(s, libsodium_dir=''):
        s.log = logging.getLogger('passprot')
        s.sodium = None
        try:
            s.LoadSodium(libsodium_dir)
            s.crypto_secretbox_KEYBYTES     = s.sodium.crypto_secretbox_keybytes()
            s.crypto_secretbox_NONCEBYTES   = s.sodium.crypto_secretbox_noncebytes()
            s.crypto_secretbox_ZEROBYTES    = s.sodium.crypto_secretbox_zerobytes()
            s.crypto_secretbox_BOXZEROBYTES = s.sodium.crypto_secretbox_boxzerobytes()
            s.crypto_pwhash_SALTBYTES       = s.sodium.crypto_pwhash_saltbytes()
            s.NONCE_LEN = s.crypto_secretbox_NONCEBYTES
        except Exception as e:
            s.load_error_msg = str(e)
            s.sodium = None

    # passes is a list of passwords. We might make it a dict with tags (quasi-usernames) later.
    # we dont actually enforce any policy on how many passes, but do all 2-combos if > 1
    # It's up to the caller to enforce stuff like "this is a root key, it needs more than 1 password" etc.

    # we pack: Salt, number of passwords used, & list of encrypted payloads.
    # Binary format: [salt sz][xpayload sz][salt][num-passes][xpayload][xpayload][xpayload]...
    # (num passes because otherwise theres no way to tell diff between 1 and 2 passwords - both product 1 payload)
    # but WE dont care about key types etc, thats the caller's problem.

    def Pack(s, num_passes, salt, xpayloads):
        plen = len(xpayloads[0])
        return encode_uvarint(num_passes) + encode_uvarint(len(salt)) + encode_uvarint(plen) + salt + b''.join(xpayloads)

    def Unpack(s, buf):
        idx = 0
        num_passes,idx   = decode_uvarint(buf, idx)
        salt_len,idx     = decode_uvarint(buf, idx)
        xpayload_len,idx = decode_uvarint(buf, idx)
        if salt_len != s.crypto_pwhash_SALTBYTES:   raise TypeError('salt is incorrect length')     # sanity
        salt = buf[idx:idx+salt_len]
        idx += salt_len
        xpayloads = [buf[i: i + xpayload_len] for i in range(idx, len(buf), xpayload_len)]
        return num_passes, salt, xpayloads


    # --- One password ---
    def SinglePassEncrypt(s, payload, passw):
        if not s.sodium:        # sodium DLL load errors on startup are surfaced here.
            raise RuntimeError(s.load_error_msg)
        if not isinstance(payload, bytes):
            raise TypeError('invalid payload for pass encrypt')
        salt = s.GetPwSalt()
        pass_key = s.PwToKey(passw, salt)
        xpayload = s.RawKeyEncrypt(pass_key, payload)
        return s.Pack(1, salt, [xpayload])

    # --- 2-password combos for N passwords ---
    def MultiPassEncrypt(s, payload, passes):
        if not isinstance(payload, bytes):  raise TypeError('invalid payload for pass encrypt')
        if not isinstance(passes, list):    raise TypeError('passes must be a list')
        if not s.sodium:
            raise RuntimeError(s.load_error_msg)
        salt = s.GetPwSalt()
        pass_keys = []
        for ia,ib in itertools.combinations(passes, 2):
            concat_pass = ''.join(sorted([ia, ib]))
            pass_keys.append(s.PwToKey(concat_pass, salt))
        xpayloads = [s.RawKeyEncrypt(pass_key, payload) for pass_key in pass_keys]
        return s.Pack(len(passes), salt, xpayloads)


    # UX flow - take buf, call DualPasswordsNeeded, use that to throw up the appropriate user forms
    #           then call SinglePassdecrypt or MultiPassDecrypt accordingly.

    def DualPasswordsNeeded(s, buf):
        num_passes,_,_ = s.Unpack(buf)
        return num_passes > 1

    def SinglePassDecrypt(s, buf, passw):
        if not s.sodium:
            raise RuntimeError(s.load_error_msg)
        nump,salt,xpayloads = s.Unpack(buf)     # only ever 1 payload rly
        pass_key = s.PwToKey(passw, salt)
        payload  = s.RawKeyDecrypt(pass_key, xpayloads[0])
        return payload

    def DualPassDecrypt(s, buf, passes):
        if not s.sodium:
            raise RuntimeError(s.load_error_msg)
        nump,salt,xpayloads = s.Unpack(buf)
        concat_pass = ''.join(sorted(passes))
        pass_key = s.PwToKey(concat_pass, salt)
        # --- try each payload in turn ---
        for xpayload in xpayloads:
            try:
                payload = s.RawKeyDecrypt(pass_key, xpayload)
                return payload
            except Exception as e:
                continue
        raise ValueError('incorrect passwords')



    # ====================== BEGIN copy pasta from socrypt ===============================
    # ====================== BEGIN copy pasta from socrypt ===============================

    def LoadSodium(s, libsodium_dir=''):
        # --- Load libsodium dll ---
        base = "libsodium"
        bit_expl = "_64" if sys.maxsize > 2 ** 32 else "_32"
        ext = {"Darwin": "dylib", "Windows": "dll", "Linux": "so"}[platform.system()]

        # dir is 1) explicit (e.g. apiweb.py bc w3wp.exe), 2) blank->current dir (dev/tool scripts), 3) appdir if frozen exe (prod EXEs)
        if not libsodium_dir and hasattr(sys, "frozen"):
            libsodium_dir = os.path.abspath(os.path.dirname(sys.executable))

        # try for _32 or _64 first, fall back to standard if they are not found.
        try_names = ["%s%s.%s" % (base, bit_expl, ext), "%s.%s" % (base, ext)]

        for try_name in try_names:
            path = Path(libsodium_dir) / try_name  # '' = current dir for libsodium_dir
            s.log.debug(" trying libsodium dll %r", path.as_posix())
            if path.is_file():
                full_path = path.resolve().as_posix()
                s.log.debug("loading libsodium dll path %r", full_path)
                s.sodium = ctypes.CDLL(full_path)
                break
        else:
            raise IOError("No libsodium dll file found")
        s.log.debug("libsodium load ok")

    def RandomBytes(s, size):
        buf = ctypes.create_string_buffer(size)
        s.sodium.randombytes(buf, ctypes.c_ulonglong(size))
        return buf.raw

    def GetPwSalt(s):
        return s.RandomBytes( s.crypto_pwhash_SALTBYTES )

    def RawKeyEncrypt(s, key, data):                    # These wont be available on HSMed CryptoBases, obviously. Currently only the PasswordAccess-er needs them i think.
        # print 'ENC key  ',repr(key)
        # print 'ENC data ',repr(data[:100])
        if not data:                                    raise ValueError('Missing data for secret_box encrypt')

        nonce   = s.RandomBytes(s.NONCE_LEN)
        padded  = (b"\x00" * s.crypto_secretbox_ZEROBYTES) + data
        out_buf = ctypes.create_string_buffer(len(padded))

        ret = s.sodium.crypto_secretbox(out_buf, padded, ctypes.c_ulonglong(len(padded)), nonce, key)

        if ret != 0:                                    raise ValueError('secret_box encrypt failed (%d)' % (ret))
        xdata = out_buf.raw[s.crypto_secretbox_BOXZEROBYTES:]
        return nonce + xdata

    def RawKeyDecrypt(s, key, xdata):
        nonce   = xdata[:s.NONCE_LEN]
        xdata   = xdata[s.NONCE_LEN:]
        if not xdata:                                   raise ValueError('Missing data for secret_box_open decrypt')

        padded  = b"\x00" * s.crypto_secretbox_BOXZEROBYTES + xdata
        out_buf = ctypes.create_string_buffer(len(padded))

        ret = s.sodium.crypto_secretbox_open(out_buf, padded, ctypes.c_ulonglong(len(padded)), nonce, key)

        # if ret != 0:                                    raise ValueError('secret_box_open decrypt failed (%d)' % (ret))
        if ret != 0:                                    raise DecryptFailed('Decrypt failed (possibly incorrect password)')
        data = out_buf.raw[s.crypto_secretbox_ZEROBYTES:]
        return data

    def PwToKey(s, password, salt):
        if len(salt) != s.crypto_pwhash_SALTBYTES:        raise ValueError("PwToKey invalid salt length")
        password = password.encode('utf-8')               # libsodium itself wants bytes, presumably

        key_out_len     = s.sodium.crypto_auth_keybytes()
        key_out_len_c   = ctypes.c_ulonglong( key_out_len                                           )
        hash_algo_c     = ctypes.c_int(       s.sodium.crypto_pwhash_alg_argon2i13()                )
        ops_limit_c     = ctypes.c_ulonglong( s.sodium.crypto_pwhash_argon2i_opslimit_interactive() )    # interactive, moderate, sensitive
        mem_limit_c     = ctypes.c_size_t(    s.sodium.crypto_pwhash_argon2i_memlimit_interactive() )
        len_password_c  = ctypes.c_ulonglong( len(password)                                         )
        out_buf         = ctypes.create_string_buffer(key_out_len)

        ret = s.sodium.crypto_pwhash(ctypes.byref(out_buf), key_out_len_c, password, len_password_c, salt, ops_limit_c, mem_limit_c, hash_algo_c)

        if ret != 0:                                    raise ValueError('crypto_pwhash failed (%d)' % (ret))
        return out_buf.raw


    # ====================== END copy pasta from socrypt ===============================
    # ====================== END copy pasta from socrypt ===============================

# Pretty sure at this point that the encrypted payloads WILL be uniform:
# if not all([plen == len(i) for i in xpayloads]):  raise ValueError('payload lengths not uniform')

def test_1_pass():
    pp = PassProtect()
    passw   = 'hello'
    payload = 'hello world'
    buf  = pp.SinglePassEncrypt(payload, passw)
    pay2 = pp.SinglePassDecrypt(buf, passw)
    assert pay2 == payload

def test_2_pass():
    pp = PassProtect()
    passes  = ['foo','bar','baz','troz','zort']
    passes2 = ['foo','baz']
    payload = 'hello world'
    buf     = pp.MultiPassEncrypt(payload, passes)
    pay2    = pp.DualPassDecrypt(buf, passes2)
    assert pay2 == payload


if __name__ == '__main__':
    #while True:
    test_2_pass()




from . import commandline

# --- setup.py entry_point for "c3" command goes via here ---
def main():
    commandline.CommandlineMain()

# --- python -m c3  commands go via here ---
if __name__ == "__main__":
    commandline.CommandlineMain()




from __future__ import unicode_literals     # for python2

import base64, traceback, random, os, datetime
from pprint import pprint

import pytest

import b3.hexdump

from c3.constants import *
from c3.errors import *
from c3.signverify import SignVerify
from c3 import structure
from c3 import commandline

@pytest.fixture
def c3m():
    c3_obj = SignVerify()
    return c3_obj

@pytest.fixture
def csr_nopass(c3m):
    ce1 = c3m.make_csr(name="harry", expiry="24 october 2024")
    ce1.private_key_set_nopassword()
    return ce1

@pytest.fixture
def ce1(c3m, csr_nopass):
    c3m.sign(csr_nopass, csr_nopass)
    return csr_nopass

@pytest.fixture
def ce1_pub_buf(c3m, ce1):
    return ce1.pub.as_binary()

@pytest.fixture
def ce1_txt(c3m, ce1):
    return ce1.both.as_text()


CERT_VIS_MAP = dict(schema=CERT_SCHEMA, field_map=["subject_name", "expiry_date", "issued_date"])
STRIP_VF = "[ Subject Name ]  harry\n[ Expiry Date  ]  24 October 2024\n[ Issued Date  ]  14 November 2022\n"

# Binary roundtrip just a CSR (pub block isn't a chain, just a cert by itself)
# turn CSR into binary, then load that, then turn THAT into binary, then check the binaries match.

def test_csr_roundtrip_binary(c3m, csr_nopass):
    ce1_bin = csr_nopass.both.as_binary()
    ce2 = c3m.load(block=ce1_bin)
    ce2_bin = ce2.both.as_binary()
    assert ce1_bin == ce2_bin

# Text roundtrip a CSR

def test_csr_roundtrip_text(c3m, csr_nopass):
    ce1_txt = csr_nopass.both.as_text()
    assert "October 2024" in ce1_txt    # ensure default Visible Fields are generated also.
    ce2 = c3m.load(text=ce1_txt)
    ce2_txt = ce2.both.as_text()
    assert ce1_txt == ce2_txt

# Remove the visible fields, ensure ce2 still loads properly and generates them.

def test_csr_roundtrip_text_strip_vf(c3m, csr_nopass):
    ce1_txt = csr_nopass.both.as_text()
    ce1_txt_noVF = ce1_txt.replace(STRIP_VF, "")
    ce2 = c3m.load(text=ce1_txt_noVF)
    ce2_txt = ce2.both.as_text()
    assert ce1_txt == ce2_txt

def test_csr_optional_cert_type(c3m, csr_nopass):
    csr = csr_nopass
    txt1 = csr.pub.as_text()
    assert "[ Cert Type" not in txt1
    csr.cert["cert_type"] = "Special cert"
    txt2 = csr.pub.as_text()
    assert "[ Cert Type"  in txt2


def test_selfsign_roundtrip_binary(c3m):
    ce1 = c3m.make_csr(name="harry", expiry="24 october 2024")
    ce1.private_key_set_nopassword()
    assert ce1.pub_type == PUB_CSR
    c3m.sign(ce1, ce1)      # self sign
    assert ce1.pub_type == PUB_CERTCHAIN
    ce1_bin = ce1.both.as_binary()

    ce2 = c3m.load(block=ce1_bin)
    ce2_bin = ce2.both.as_binary()
    assert ce1_bin == ce2_bin


def test_ss_verify_binary(c3m):
    ce1 = c3m.make_csr(name="harry", expiry="24 october 2024")
    ce1.private_key_set_nopassword()
    c3m.sign(ce1, ce1)
    ce1_bin = ce1.both.as_binary()

    c3m.load_trusted_cert(block=ce1_bin)
    ce2 = c3m.load(block=ce1_bin)
    assert c3m.verify(ce2) is True

def test_ss_verify_text(c3m):
    ce1 = c3m.make_csr(name="harry", expiry="24 october 2024")
    ce1.private_key_set_nopassword()
    c3m.sign(ce1, ce1)
    ce1_txt = ce1.both.as_text()
    assert "Expiry Date  ]" in ce1_txt

    c3m.load_trusted_cert(text=ce1_txt)
    ce2 = c3m.load(text=ce1_txt)
    assert c3m.verify(ce2) is True

# --- Try to sign with expired cert ---
def test_sign_expired(c3m):
    expi = c3m.make_csr(name="expi", expiry="24 oct 2002")
    with pytest.raises(CertExpired):
        c3m.sign(expi, expi)

# ----- Inter cert signing / verifying ----

# Note: CEs must come in via load() for full chain-unpacking, dont use them directly.

def test_inter_sign_verify(c3m):
    selfsigned = c3m.make_csr(name="root1", expiry="24 october 2024")
    c3m.sign(selfsigned, selfsigned)

    inter = c3m.make_csr(name="inter2", expiry="24 oct 2024")
    assert inter.pub_type == PUB_CSR
    c3m.sign(inter, selfsigned)
    assert inter.pub_type == PUB_CERTCHAIN

    c3m.load_trusted_cert(block=selfsigned.pub.as_binary())
    inter2 = c3m.load(block=inter.pub.as_binary())
    assert c3m.verify(inter2) is True

# ----- verify-chain metadata functions ----

def test_inter_namechain(c3m):
    root1 = c3m.make_csr(name="root1", expiry="24 october 2024", cert_type="rootcert")
    c3m.sign(root1, root1)
    inter2 = c3m.make_csr(name="inter2", expiry="24 oct 2024", cert_type="inter")
    c3m.sign(inter2, root1, link_by_name=True)
    inter3 = c3m.make_csr(name="inter3", expiry="24 oct 2024", cert_type="inter")
    c3m.sign(inter3, inter2, link_by_name=False)
    inter4 = c3m.make_csr(name="inter4", expiry="24 oct 2024", cert_type="inter")
    c3m.sign(inter4, inter3, link_by_name=False)
    pay = c3m.make_payload(b"hello world")
    c3m.sign(pay, inter4, link_by_name=False)
    # --- Verify ---
    pay_block = pay.pub.as_binary()
    pay2 = c3m.load(block=pay_block)

    c3m.load_trusted_cert(block=root1.pub.as_binary())
    # c3m.load_trusted_cert(block=inter2.pub.as_binary())
    # c3m.load_trusted_cert(block=inter3.pub.as_binary())
    c3m.verify(pay2)
    assert pay2.vtypes() == "/rootcert/inter/inter/inter"
    assert pay2.vnames() == "/root1/inter2/inter3/inter4"


# ----- Payload signing / verifying ----

def test_payload_sign_verify(c3m):
    selfsigned = c3m.make_csr(name="root1", expiry="24 october 2024")
    c3m.sign(selfsigned, selfsigned)

    payload = b"Hello i am a payload"
    pce = c3m.make_payload(payload)
    assert pce.pub_type == BARE_PAYLOAD
    c3m.sign(pce, selfsigned)
    assert pce.pub_type == PUB_PAYLOAD

    c3m.load_trusted_cert(block=selfsigned.pub.as_binary())
    pce2 = c3m.load(block=pce.pub.as_binary())
    assert c3m.verify(pce2) is True

# ---- Sign using intermediate ----

def test_payload_sign_intermediate(c3m):
    selfsigned = c3m.make_csr(name="root1", expiry="24 october 2024")
    c3m.sign(selfsigned, selfsigned)
    inter = c3m.make_csr(name="inter2", expiry="24 oct 2024")
    c3m.sign(inter, selfsigned)
    payload = b"Hello i am a payload"
    pce = c3m.make_payload(payload)
    c3m.sign(pce, inter)
    pce_bin = pce.pub.as_binary()
    # --------------------------------------------------------------
    c3m.load_trusted_cert(block=selfsigned.pub.as_binary())
    pce2 = c3m.load(block=pce_bin)
    assert c3m.verify(pce2) is True




# --- load-to-sign (instead of make_csr-to-sign ---

def test_load_to_sign(c3m):
    selfsigned = c3m.make_csr(name="root1", expiry="24 october 2024")
    c3m.sign(selfsigned, selfsigned)
    selfsigned.private_key_set_nopassword()
    ss2 = c3m.load(block=selfsigned.both.as_binary())
    inter = c3m.make_csr(name="inter2", expiry="24 oct 2024")
    c3m.sign(inter, ss2)

# --- Didn't set the priv key bare (or encrypt) ---

def test_load_to_sign_priv_key_unset(c3m):
    selfsigned = c3m.make_csr(name="root1", expiry="24 october 2024")
    c3m.sign(selfsigned, selfsigned)
    with pytest.raises(OutputError):
        ss_bin = selfsigned.both.as_binary()


# ---- Private key encrypt/decrypt (password in code) ---

def test_privkey_encrypt(c3m):
    selfsigned = c3m.make_csr(name="root1", expiry="24 october 2024")
    c3m.sign(selfsigned, selfsigned)
    selfsigned.private_key_encrypt("hunter3")

    ss2 = c3m.load(block=selfsigned.both.as_binary())
    ss2.private_key_decrypt("hunter3")
    inter = c3m.make_csr(name="inter2", expiry="24 oct 2024")

    c3m.sign(inter, ss2)


def test_privkey_encrypt_env_var(c3m):
    selfsigned = c3m.make_csr(name="root1", expiry="24 october 2024")
    c3m.sign(selfsigned, selfsigned)
    os.environ["C3_PASSWORD"] = "Password01!"   # Note: this side-effects the rest of the tests!
    selfsigned.private_key_encrypt_user()

    ss2 = c3m.load(block=selfsigned.both.as_binary())
    ss2.private_key_decrypt_user()
    inter = c3m.make_csr(name="inter2", expiry="24 oct 2024")

    c3m.sign(inter, ss2)

# glitch a privkey byte to exercise the integrity check
def test_privkey_bare_integrity(c3m):
    bare_priv = b"hello world"
    priv_block_bytes = structure.make_priv_block(bare_priv, bare=True)
    priv_block_bytes = priv_block_bytes[:16] + b"a" + priv_block_bytes[17:]
    with pytest.raises(IntegrityError):
        structure.load_priv_block(priv_block_bytes)


# ----- Verification errors -----

# Glitch the payload contents so the signature fails to verify
def test_verify_signature_fail(c3m, ce1_pub_buf):
    c3m.load_trusted_cert(block=ce1_pub_buf)
    buf = ce1_pub_buf[:150] + b"X" + ce1_pub_buf[151:]
    ce2 = c3m.load(block=buf)
    with pytest.raises(InvalidSignatureError):
        c3m.verify(ce2)

# Apart from actual signature fails, there are 3 ways for this to fail:
# 1) "fell off" - unnamed issuer cert and no next cert in line ("short chain")
# 2) Cant Find Named Cert - in the cert store / trust store / certs_by_name etc
# 3) Last cert is self-signed and verified OK but isn't in the trust store.


# cut the signing cert off of the end of intermediate cert public block
# to trigger "next cert is the signer but there is no next cert" failure mode
# note: Don't need trusted root1 loaded because it doesn't get that far
inter_text = """
2Q7uAukZsAEJAGUJAAVpbnRlchkBBWludGVyOQIBAQkDQJqwMvo4wnp9Auz93zKrscBlbu8mihvz
RARudJyYDumHVutovutOUEaPiJuQjMpI/vOG+hxn6uhZv7/aTM+B+e+pBAQYCtAfqQUEFQvMHwkB
RQkAQMmN9FKivxybOhv+dkqU2qhhcr3DSSvKqdFF4jBBp3yjV4mT3JJ1mdJ2m8sJq9cIyeRAhzgF
kBJD7wBH/RnwrNQBAekZtgEJAGUJAAVyb290MRkBBXJvb3QxOQIBAQkDQDiZmemfIXf+5WpBaAJY
OlgqitfC4txC5kBdGEgaA0A+jblvdn5LdxdO0heNDDNBh9nYDZUJUwo+nOpyd/H05kWpBAQYCtAf
qQUEFQvMHwkBSwkAQEbrVdKzH6YpeQiKo9bKmHWUV0yIu6H4tsOfrGJhEnARutowDamrPlcBdT9k
xqzUkauFU6Dd+xyEFoYaLEOTrJ4JAQVyb290MQ=="""
inter_block = base64.b64decode(inter_text)

def test_verify_short_chain(c3m):
    buf = inter_block[:]
    buf2 = buf[:184]
    inter = c3m.load(block=buf2)
    with pytest.raises(ShortChainError):
        c3m.verify(inter)

# try verify an intermediate that has a 'by name' cert link (instead of append-it-in),
# to induce CertNotFoundError. (by just not loading a trusted cert first.)
interbyname_text = """
2Q68AekZuAEJAGcJAAZpbnRlcjIZAQZpbnRlcjI5AgEBCQNAoRCrLvSQ/TBs0KBamYcu7j++Mw73
Hfa/9IuF2eQO9HwAibgjnUsuf+U+fGFKzx/hzd1zvWZJD9Dc61nnX/qVRakEBBgK0B+pBQQVC8wf
CQFLCQBAEWjuOWLDXzfkAK1eSUmtSrXsPzGtWfegHSGg/jbr4g92omM4d26rxj4BsslGXSp1ysHX
3vLXaCweTEGYLU82yQkBBXJvb3Qx"""
interbyname_block = base64.b64decode(interbyname_text)

def test_verify_cert_not_found(c3m):
    interbyname = c3m.load(block=interbyname_block)
    with pytest.raises(CertNotFoundError):
        c3m.verify(interbyname)


# Without loading ce1 to trusted store first,
# a fully valid chain with a selfsign at the end, should fail with UntrustedChainError
def test_verify_untrusted_chain_error(c3m, ce1_pub_buf):
    ce2 = c3m.load(block=ce1_pub_buf)
    with pytest.raises(UntrustedChainError):
        c3m.verify(ce2)

# ---- Quick fuzz to exercise all the load errors ----
# just throw all the error strings together then assert the presence of the errors we want

selfsigned_pub_and_priv_b64 = """
6Qv7AQkAvgHZDroB6Rm2AQkAZQkABXJvb3QxGQEFcm9vdDE5AgEBCQNAOJmZ6Z8hd/7lakFoAlg
6WCqK18Li3ELmQF0YSBoDQD6NuW92fkt3F07SF40MM0GH2dgNlQlTCj6c6nJ38fTmRakEBBgK0B
+pBQQVC8wfCQFLCQBARutV0rMfpil5CIqj1sqYdZRXTIi7ofi2w5+sYmEScBG62jANqas+VwF1P
2TGrNSRq4VToN37HIQWhhosQ5OsngkBBXJvb3QxCQE26Q8zOQABATkBAQEJAiDVJDmswBXrxa4P
IuqHpl1H38C5Q1QdssQSxPhUh1ZAIzkDBb3RuKgB
"""
# note: verified there are no \xaa bytes in the above block.
# Note: The above block was created using PRIME256V1 keys not SECP256k1, so it serves as a test for
#       keypairs.py's multi-keytype support also.

# Note: There are some header data_lens which are not actually used by us. So when those get
#       glitched, the verify still succeeds, which is legit and ok.
#       Cross reference printing Success! in here with printing data_len in expect_key_header()
#       in structure.py to prove this out.

def test_load_fuzz(c3m):
    buf = base64.b64decode(selfsigned_pub_and_priv_b64)
    c3m.load_trusted_cert(block=buf)
    fails = set()
    for i in range(len(buf)):
        buf2 = buf[:i] + b"\xaa" + buf[i+1:]
        try:
            ce = c3m.load(block=buf2)
            c3m.verify(ce)
            # print("Success! i=", i)
        except Exception as e:
            fails.add(str(e))

    errs = "\n".join(fails)
    # print(errs)
    assert "List item header structure is invalid" in errs
    assert "List item header key invalid" in errs
    # assert "invalid - no data" in errs
    # assert "data is missing" in errs
    assert "Mandatory field" in errs
    # assert "Required schema field" in errs
    assert "Private key block failed data integrity check" in errs
    assert "Signature failed to verify" in errs
    assert "not found" in errs
    assert "Next issuer cert is missing" in errs
    assert "Incorrect key in header" in errs



# ---- Test keypair matching ----

# We have to hack inner CE structures, thanks to the CE system, mismatches cant happen as a result
# of api misuse / user error.

def test_keypair_matching(c3m):
    ce1 = c3m.make_csr(name="root1", expiry="24 oct 2024")
    ce2 = c3m.make_csr(name="root1", expiry="24 oct 2024")  # note simulating same name
    ce2.priv_key_bytes = ce1.priv_key_bytes   # overwrite ce2 priv with ce1 priv.
    with pytest.raises(SignError, match="Private key and public key do not match"):
        c3m.sign(ce2, ce2)



# ---- Test load error handling ----

def test_load_empty(c3m):
    with pytest.raises(ValueError):
        c3m.load(block=b"")     # fails highlander_check

def test_load_none(c3m):
    with pytest.raises(ValueError):
        c3m.load(block=None)    # fail highlander_check

def test_load_nulls(c3m):
    with pytest.raises(StructureError):     # fails initial expect_header
        c3m.load(block=b"\x00\x00\x00\x00\x00\x00\x00\x00")



# --- Visible fields tamper checks ---
def test_vis_busted_vertical(c3m, ce1_txt):
    busted_vertical_structure = "\n\n".join(ce1_txt.splitlines())
    with pytest.raises(StructureError, match="structure is invalid"):
        c3m.load(text=busted_vertical_structure)

def test_vis_bad_field(c3m, ce1_txt):
    bad_vf = ce1_txt.replace("Date  ]", "Date]")
    with pytest.raises(TamperError, match="format for visible"):
        c3m.load(text=bad_vf)

def test_vis_spurious_field(c3m, ce1_txt):
    spurious_field = "[ Spurious Field ]  Hello world"
    bad_vf = ce1_txt.replace("[ Subject Name ]  harry", spurious_field)
    with pytest.raises(TamperError, match="not present in the secure area"):
        c3m.load(text=bad_vf)

def test_vis_value_mismatch(c3m, ce1_txt):
    spurious_field = "[ Subject Name ]  Harold"
    bad_vf = ce1_txt.replace("[ Subject Name ]  harry", spurious_field)
    with pytest.raises(TamperError, match="does not match secure"):
        c3m.load(text=bad_vf)



# ----- Visible fields for custom payloads -----
LI_SCHEMA = (
    (b3.UTF8, "typ", 0, True),   # "License type"
    (b3.UTF8, "org", 4, False),  # "Organization"
    (b3.UTF8, "hostnames", 5, False),  # "Hostnames"
)
LI_VISFIELDS = [ ["org", "Organization"], "hostnames", ["typ", "License Type"] ]
LI_VISMAP = dict(schema=LI_SCHEMA, field_map=LI_VISFIELDS)

def test_payload_verify_text(c3m):
    selfsigned = c3m.make_csr(name="root1", expiry="24 october 2024")
    c3m.sign(selfsigned, selfsigned)
    payload_d = dict(typ="type 1", org="Hello Ltd", hostnames="fred")
    payload = b3.schema_pack(LI_SCHEMA, payload_d)
    pce = c3m.make_payload(payload)
    c3m.sign(pce, selfsigned)
    pce_txt = pce.pub.as_text()
    c3m.load_trusted_cert(block=selfsigned.pub.as_binary())
    pce2 = c3m.load(text=pce_txt)
    assert c3m.verify(pce2) is True
    assert pce2.payload == payload


def test_payload_verify_text_visfields_noschema(c3m):
    selfsigned = c3m.make_csr(name="root1", expiry="24 october 2024")
    c3m.sign(selfsigned, selfsigned)
    payload_d = dict(typ="type 1", org="Hello Ltd", hostnames="fred")
    payload = b3.schema_pack(LI_SCHEMA, payload_d)
    pce = c3m.make_payload(payload)
    c3m.sign(pce, selfsigned)
    pce_txt = pce.pub.as_text(vis_map=LI_VISMAP)
    c3m.load_trusted_cert(block=selfsigned.pub.as_binary())
    # Load needs to be supplied the vis_map, if there are visible fields incoming,
    # otherwise load can't tamper-check the visible fields.
    with pytest.raises(StructureError, match="schema unknown"):
        c3m.load(text=pce_txt)


def test_payload_verify_text_visfields(c3m):
    selfsigned = c3m.make_csr(name="root1", expiry="24 october 2024")
    c3m.sign(selfsigned, selfsigned)
    payload_d = dict(typ="type 1", org="Hello Ltd", hostnames="fred")
    payload = b3.schema_pack(LI_SCHEMA, payload_d)
    pce = c3m.make_payload(payload)
    c3m.sign(pce, selfsigned)
    pce_txt = pce.pub.as_text(vis_map=LI_VISMAP)
    c3m.load_trusted_cert(block=selfsigned.pub.as_binary())
    pce2 = c3m.load(text=pce_txt, vis_map=LI_VISMAP)
    assert c3m.verify(pce2) is True


def test_payload_verify_text_visfields_tamper(c3m):
    selfsigned = c3m.make_csr(name="root1", expiry="24 october 2024")
    c3m.sign(selfsigned, selfsigned)
    payload_d = dict(typ="type 1", org="Hello Ltd", hostnames="fred")
    payload = b3.schema_pack(LI_SCHEMA, payload_d)
    pce = c3m.make_payload(payload)
    c3m.sign(pce, selfsigned)
    pce_txt = pce.pub.as_text(vis_map=LI_VISMAP)
    pce_txt = pce_txt.replace("fred", "albert")      # change visible field value
    c3m.load_trusted_cert(block=selfsigned.pub.as_binary())
    with pytest.raises(TamperError, match="fred"):
        c3m.load(text=pce_txt, vis_map=LI_VISMAP)



# --- Commandline testing ---

# Note: this writes files to folder!
PAYLOAD = "Hello i am a payload\n"
COMMAND_LINES = """
c3 make        --name=root1  --expiry="24 oct 2024" --parts=split
c3 signcert    --name=root1  --using=self           --parts=split
c3 make        --name=inter1 --expiry="24 oct 2024" --parts=combine
c3 signcert    --name=inter1 --using=root1          --parts=combine
c3 signpayload --payload=payload.txt --using=inter1
c3 verify      --name=payload.txt --trusted=root1
"""

#@pytest.mark.skip(reason="writes files - enable in code for more coverage")
def test_commandline_full():
    os.environ["C3_PASSWORD"] = "Password01!"
    with open("payload.txt", "wt") as f:
        f.write(PAYLOAD)
    for line in COMMAND_LINES.splitlines():
        if not line.strip():
            continue
        commandline.CommandlineMain(line)
    # not really a proper test, should assert something.




# ---------- ID collision ---------------

# todo: port this test at some point. It's slightly dependend on ID behaviour.

# Note signing doesnt fail when we *append* the root9 cert itself into the chain
#      which you're not supposed to do. It succeeds because root9 is in trusted_CEs and verify
#      sees that the NAME root9 is in trusted_CEs so sets the found_in_trusted flag so that
#      UntrustedChainError doesn't trigger at the end.

# Note this looks like it would open us up to malicious actors appending their own cert with the same
#      ID, but the actual signature verification step is always done, which defends against this,
#      as shown by the next test.

# def test_sign_rootcert_namecollide(c3m):
#     expir = datetime.date(2023, 9, 9)
#     # Legit guy
#     root_pub, root_priv = c3m.make_sign(MAKE_SELFSIGNED, name="root5", expiry=expir)
#     c3m.add_trusted_certs(root_pub)
#     # Attacker guy
#     evil_pub, evil_priv = c3m.make_sign(MAKE_SELFSIGNED, name="root5", expiry=expir)   # NOTE same name
#     # evil chain
#     inter_pub, inter_priv = c3m.make_sign(MAKE_INTERMEDIATE, name="inter9", using_pub=evil_pub, using_priv=evil_priv, expiry=expir)
#     chain = structure.load_pub_block(inter_pub)
#     with pytest.raises(InvalidSignatureError):
#         ret = c3m.verify(chain)
#




# C3 Command-line operations (make keys, sign etc).
# Doubles as usage examples for the various components.

from __future__ import print_function

import sys, re, datetime, shlex
from pprint import pprint

from c3.signverify import SignVerify
from c3.errors import NoPassword

# Use cases:
# * Make license (sign),  verify license
# * Make acmd key, make (sign) acmd message, verify acmd message
# * make build key [nocode], sign build manifest [nocode], verify build manifest  [signverify]

def CommandlineMain(cmdline_str=""):
    args = ArgvArgs(cmdline_str)    # cmdline_str is for testing, usually this pulls from sys.argv
    cmd = args.cmd
    c3m = SignVerify()

    try:
        # --- CSR / certchain pipeline ---
        if cmd == "make":
            # --- pub cert (signing request) ---
            csr = c3m.make_csr(name=args.name, expiry=args.expiry, cert_type=args.type)
            # --- private key (encrypt) ---
            if "nopassword" in args:
                csr.private_key_set_nopassword()
            else:
                try:
                    csr.private_key_encrypt_user()
                except NoPassword:
                    print("\nNo password entered, bailing.")
                    return
            # --- save file(s) ---
            csr.write_files(args.parts, True)     # note: args.parts can be None
            print("Success!")
            return

        if cmd == "signcert":
            to_sign = c3m.load(filename=args.name)
            if args.using == "self":
                signer = to_sign
            else:
                signer = c3m.load(filename=args.using)
            link_by_name = "link" in args and args.link == "name"
            signer.private_key_decrypt_user()
            c3m.sign(to_sign, signer, link_by_name)
            to_sign.write_files(args.parts, True)
            print("Success!")
            return

        # --- Payload pipeline ---
        if cmd == "signpayload":
            signer = c3m.load(filename=args.using)
            payload = c3m.make_payload(open(args.payload, "rb").read())
            signer.private_key_decrypt_user()
            c3m.sign(payload, signer)
            payload.write_prints = True
            payload.pub.write_text_file(args.payload)
            print("Success!")
            return

        # --- Load & verify ---
        if cmd == "verify":
            c3m.load_trusted_cert(filename=args.trusted)
            ce = c3m.load(filename=args.name)
            if c3m.verify(ce):
                print("\nVerify OK")
            return

        if cmd == "load":
            x = c3m.load(filename=args.name)
            print("pub_type ", x.pub_type)
            print("chain    ")
            pprint(x.chain)
            print("payload  ")
            print(x.payload)
            return

        Usage()
        print("Unknown Command %r" % cmd)

    except Exception as e:
        if "debug" in args:
            raise
        else:
            Usage()
            print("ERROR:  "+str(e))
            return


def Usage(msg=""):
    help_txt = """%s
Usage:
    c3 make        --name=root1  --expiry="24 oct 2024" 
    c3 signcert    --name=root1  --using=self           
    c3 make        --name=inter1 --expiry="24 oct 2024"
    c3 signcert    --name=inter1 --using=root1
    c3 signpayload --payload=payload.txt --using=inter1
    c3 verify      --name=payload.txt    --trusted=root1
    make options   --type=rootcert --parts=split/combine --nopassword=y
    sign options   --link=name/append
    """ % (msg,)
    print(help_txt)

class ArgvArgs(dict):
    def __init__(self, cmdline_str=""):
        super(ArgvArgs, self).__init__()
        if cmdline_str:                         # for testing
            argv = shlex.split(cmdline_str)
        else:
            argv = sys.argv
        if len(argv) < 2:
            Usage()
            sys.exit(1)
        self.cmd = argv[1].strip().lower()
        for arg in argv:
            z = re.match(r"^--(\w+)=(.+)$", arg)
            if z:
                k, v = z.groups()
                self[k] = v
        self.optional_args = ["type", "parts"]
    def __getattr__(self, name):
        if name not in self:
            if name in self.optional_args:
                return None
            raise Exception("Please specify missing commandline argument   --%s="%name)
        return self[name]


